/*
 * Copyright 2005-2007 Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "xf86.h"
#include "xf86_OSproc.h"

#include "xf86fbman.h"

#include "via_driver.h"
#include "via_memory.h"

/*
 *
 */
void
ViaMemInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
{
    VIAPtr pVia = VIAPTR(pScrn);
    unsigned long offset, size;
    CARD16 Bpp = pScrn->bitsPerPixel >> 3;

    VIAFUNC(pScrn);

    if (!pVia->FBOffscreenStart || !pVia->FBOffscreenSize)
	return;

    ViaDebug(pScrn->scrnIndex, "%s: FB %p (%08X) is available for allocation.\n",
             __func__, pVia->FBFreeStart, pVia->FBFreeSize);

    /* Why doesn't the FB manager handle alignment itself? */
    offset = (pVia->FBOffscreenStart + Bpp - 1 ) / Bpp;
    size = pVia->FBOffscreenSize / Bpp - offset;

    if (size > 0)
        xf86InitFBManagerLinear(pScreen, offset, size);
}

/*
 *
 */
static void
ViaMemFBFree(ScrnInfoPtr pScrn, struct ViaMem *Mem)
{
    VIAFUNC(pScrn);

    xf86FreeOffscreenLinear((FBLinearPtr) Mem->Private);

    /* sick bastards manually free this Private,
       normal people know better */
}

/*
 *
 */
static struct ViaMem *
ViaMemFBAlloc(ScrnInfoPtr pScrn, unsigned long size, CARD8 alignment)
{
    struct ViaMem *Mem;
    FBLinearPtr linear;
    unsigned long tempsize;
    CARD16 Bpp = pScrn->bitsPerPixel >> 3;

    VIAFUNC(pScrn);

    /* Make sure we don't truncate requested size */
    tempsize = (size + alignment + Bpp - 1) / Bpp;

    linear = xf86AllocateOffscreenLinear(pScrn->pScreen, tempsize, 32,
                                         NULL, NULL, NULL);
    if (!linear)
	return NULL;

    Mem = xnfcalloc(sizeof(struct ViaMem), 1);
    Mem->Base = linear->offset * Bpp;

    /* adjust for alignment */
    if (alignment) {
        alignment -= 1;
        Mem->Base = (Mem->Base + alignment) & ~alignment;
    }

    Mem->Size = size;

    Mem->Private = linear;
    Mem->PrivateSize = sizeof(FBLinear);

    Mem->Free = ViaMemFBFree;

    return Mem;
}

#ifdef HAVE_DRI
/*
 * DRM.
 *
 */

/* First some workarounds for the badly trackable drm changes.
 * Plainly using VIDEO as a macro is a pretty bad initial choice
 * though */
#ifndef VIDEO
#define VIDEO VIA_MEM_VIDEO
#endif

/*
 *
 */
static void
ViaMemDRMFree(ScrnInfoPtr pScrn, struct ViaMem *Mem)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    if (drmCommandWrite(pVia->drmFD, DRM_VIA_FREEMEM, (drm_via_mem_t *) Mem->Private,
                        sizeof(drm_via_mem_t)) < 0)
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: Failed to deallocate Memory.\n",
                   __func__);

    free((drm_via_mem_t *) Mem->Private);
}

/*
 * Trick: Allocate size + alignment, and then align Base.
 *
 */
static struct ViaMem *
ViaMemDRMAlloc(ScrnInfoPtr pScrn, unsigned long size, CARD8 alignment)
{
    VIAPtr pVia = VIAPTR(pScrn);
    struct ViaMem *Mem;
    drm_via_mem_t * DrmMem;
    int ret;

    VIAFUNC(pScrn);

    DrmMem = xnfcalloc(sizeof(drm_via_mem_t), 1);

    DrmMem->context = 1;
    DrmMem->size = size + alignment;
    DrmMem->type = VIDEO;

    ret = drmCommandWriteRead(pVia->drmFD, DRM_VIA_ALLOCMEM, DrmMem,
                              sizeof(drm_via_mem_t));

    if (ret || (DrmMem->size != (size + alignment))) {
        free(DrmMem);
        return NULL;
    }

    Mem = xnfcalloc(sizeof(struct ViaMem), 1);

    if (alignment) {
        alignment -= 1;
        Mem->Base = (DrmMem->offset + alignment) & ~alignment;
    } else
        Mem->Base = DrmMem->offset;
    Mem->Size = size;

    Mem->Private = DrmMem;
    Mem->PrivateSize = sizeof(drm_via_mem_t);

    Mem->Free = ViaMemDRMFree;

    return Mem;
}

#endif /* HAVE_DRI */


/*
 * General functions.
 *
 */

/*
 *
 */
struct ViaMem *
ViaMemAlloc(ScrnInfoPtr pScrn, unsigned long size, CARD8 alignment)
{
    struct ViaMem *Mem;

    VIAFUNC(pScrn);

    if (!size) {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "%s: libv is an idiot"
                   " (IQ %ld).\n", __func__, size);
        return NULL;
    }

#ifdef HAVE_DRI
    if (VIAPTR(pScrn)->directRenderingEnabled)
        Mem = ViaMemDRMAlloc(pScrn, size, alignment);
    else
#endif /* HAVE_DRI */
        Mem = ViaMemFBAlloc(pScrn, size, alignment);

    if (!Mem)
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "%s: Allocation failed "
                   "(%ldbytes requested).\n", __func__, size);
    else
        ViaDebug(pScrn->scrnIndex, "Allocated %d bytes at %p.\n", Mem->Size,
                 Mem->Base);
    return Mem;
}

/*
 *
 */
void
ViaMemFree(ScrnInfoPtr pScrn, struct ViaMem *Mem)
{
    VIAFUNC(pScrn);

    if (Mem) {
        ViaDebug(pScrn->scrnIndex, "Freeing %d bytes at %p.\n", Mem->Size,
                 Mem->Base);

        if (Mem->Free)
            Mem->Free(pScrn, Mem);

        free(Mem);
    }
}
