/*
 * Copyright 2006-2009 Luc Verhaegen <libv@skynet.be>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <xf86.h>

#include "via_driver.h"
#include "via_mpeg.h"
#include "via_mpegregs.h"

/*
 *
 * MPEG engine.
 *
 */

/*
 *
 */
static void
ViaMpegReset(struct ViaMpeg *Mpeg)
{
    int i, j;

    for (i = 0; i < 14; i++)
	Mpeg->Regs->Null = 0;

    Mpeg->Regs->SliceControl3 = 0x00400000; /* reset slice engine */

    for (i = 0; i < 6; i++) {
	Mpeg->Regs->Control = 0x00000063;
	for (j = 0; j < 4; j++)
            Mpeg->Regs->MV[j] = 0;
    }

    Mpeg->Regs->Control = 0x000000E3;
    for (j = 0; j < 4; j++)
        Mpeg->Regs->MV[j] = 0;
}

/*
 *
 */
void
ViaMpegEngineIdle(struct ViaMpeg *Mpeg)
{
    int i;

    for (i = 0; i < 0x10000; i++)
	if (!(Mpeg->Regs->Status & 0x00000001)) {
	    if (i > 0x2000)
		ViaDebug(Mpeg->scrnIndex, "%s: %d loops\n", __func__, i);
	    return;
	}

    xf86DrvMsg(Mpeg->scrnIndex, X_WARNING, "%s: Timed out! (0x%08X)\n",
	       __func__, (unsigned int) Mpeg->Regs->Status);
}

/*
 *
 */
Bool
ViaMpegEngineInit(struct ViaMpeg *Mpeg, CARD16 Width, CARD16 Height)
{
    VIAFUNC(Mpeg);

    /* Since MPEG macroblocks are 16x16 anyway, the chroma planes are always
       addressable by the unichrome. */
    if ((Width & 0x0F) || (Height & 0x0F)) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING, "%s: %dx%d is not a multiple of"
		   " 16x16.\n", __func__, Width, Height);
	return FALSE;
    }

    Mpeg->MBWidth = Width >> 4;
    Mpeg->MBHeight = Height >> 4;
    Mpeg->Width = (Width + 0x1F) & ~0x1F;
    Mpeg->Height = Height;

    ViaDebug(Mpeg->scrnIndex, "%s: %dx%d (MB %dx%d)\n", __func__,
	     Mpeg->Width, Mpeg->Height, Mpeg->MBWidth, Mpeg->MBHeight);

    /* start with a clean slate and fix possible previous hangs */
    ViaMpegReset(Mpeg);

    return TRUE;
}

/*
 *
 */
Bool
ViaMpegBufferGrab(struct ViaMpeg *Mpeg, XID BufferID)
{
    int i, size;
    CARD8 *Base;

    VIAFUNC(Mpeg);

    for (i = 0; i < VIAMPEG_BUFFERS_MAX; i++)
        if (!Mpeg->Buffers[i].ID && !Mpeg->Buffers[i].Mem)
	    break;

    if (i == VIAMPEG_BUFFERS_MAX) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING, "%s: No slots available.\n", __func__);
	return FALSE;
    }

    size = (Mpeg->Width * Mpeg->Height * 3) / 2;

    Mpeg->Buffers[i].Mem = ViaMemAlloc(xf86Screens[Mpeg->scrnIndex], size, 16);
    if (!Mpeg->Buffers[i].Mem) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING, "%s: Failed to allocate %dx%d buffer.\n",
		   __func__, Mpeg->Width, Mpeg->Height);
	return FALSE;
    }

    Mpeg->Buffers[i].ID = BufferID;

    Base = VIAPTR(xf86Screens[Mpeg->scrnIndex])->FBBase + Mpeg->Buffers[i].Mem->Base;
    memset(Base, 0x10, Mpeg->Width * Mpeg->Height);
    memset(Base + Mpeg->Width * Mpeg->Height, 0x80, Mpeg->Width * Mpeg->Height / 2);

    ViaDebug(Mpeg->scrnIndex, "%s: 0x%08X: Grabbed Buffer %d (%dbytes)\n",
	     __func__, BufferID, i, size);

    return TRUE;
}

/*
 *
 */
void
ViaMpegBufferRelease(struct ViaMpeg *Mpeg, XID BufferID)
{
    int i;

    VIAFUNC(Mpeg);

    /* If the buffer is still used by the decoder, first
       wait until the decoder is done. */
    if ((BufferID == Mpeg->CurrentID) ||
	(BufferID == Mpeg->FRefID) ||
	(BufferID == Mpeg->BRefID)) {
        ViaMpegEngineIdle(Mpeg);
	Mpeg->CurrentID = 0;
	Mpeg->FRefID = 0;
	Mpeg->BRefID = 0;
    }

    /* don't overwrite a picture while it's still flipping */
    if (BufferID == Mpeg->FlippingID) {
        ViaHQVIdle(VIAPTR(xf86Screens[Mpeg->scrnIndex])->Swov);
        Mpeg->FlippingID = 0;
    }

    for (i = 0; i < VIAMPEG_BUFFERS_MAX; i++)
	if (Mpeg->Buffers[i].ID == BufferID)
	    break;

    if (i == VIAMPEG_BUFFERS_MAX) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Buffer 0x%08X doesn't exist.\n",
		   __func__, (unsigned int) BufferID);
	return;
    }

    ViaDebug(Mpeg->scrnIndex, "%s: Releasing Buffer 0x%08X\n", __func__,
	     (unsigned int) BufferID);

    Mpeg->Buffers[i].ID = 0;
    if (Mpeg->Buffers[i].Mem)
	ViaMemFree(xf86Screens[Mpeg->scrnIndex], Mpeg->Buffers[i].Mem);
    Mpeg->Buffers[i].Mem = NULL;
}

/*
 *
 */
void
ViaMpegQMatrix(struct ViaMpeg *Mpeg, CARD32 Type, CARD8 Matrix[64])
{
    CARD32 *Temp = (CARD32 *) Matrix;
    int i;

    MPEGDebug(VIAFUNC(Mpeg));

    ViaMpegEngineIdle(Mpeg);

    Mpeg->Regs->QMatrixSelect = Type & 0x03;

    for (i = 0; i < 16; i++)
        Mpeg->Regs->QMatrixContent = Temp[i];
}

/*
 *
 */
int
ViaMpegSliceInit(struct ViaMpeg *Mpeg, XID BufferID, XID FRefID, XID BRefID,
		 CARD8 F_MV_Range, CARD8 B_MV_Range, CARD8 PictureStructure,
		 CARD8 FrameType, Bool ScanningAlternate, Bool PredictionFrame,
		 Bool TopFieldFirst, Bool IntraHasMV, Bool UseQScaleTableType1,
		 CARD8 IntraDCPrecision, Bool UseIntraDCTable1)
{
    CARD32 Base, CbOffset, CrOffset, Offset = 0;
    CARD32 Picture = 0, SliceControl1 = 0, SliceControl2 = 0, SliceControl3 = 0;
    CARD16 Width, Height;
    int Buffer = 0xFFFFFFFF, FRef = 0xFFFFFFFF, BRef = 0xFFFFFFFF;

    /* TODO: we need to adjust Width and Height when using an interlaced stream */
    MPEGDebug(ViaDebug(Mpeg->scrnIndex, "%s: 0x%08X: FRef 0x%02X; BRef 0x%02X\n",
		       __func__, BufferID, FRefID, BRefID));

    /* Check ID */
    for (Buffer = 0; Buffer < VIAMPEG_BUFFERS_MAX; Buffer++)
	if (Mpeg->Buffers[Buffer].ID == BufferID)
	    break;

    if (Buffer == VIAMPEG_BUFFERS_MAX) {
        xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Wrong Buffer ID (0x%08X).\n",
		   __func__, (unsigned int) BufferID);
        return BadAccess;
    }

    Mpeg->CurrentID = BufferID;
    Width = Mpeg->Width;
    Height = Mpeg->Height;

    /* Check Forward reference */
    for (FRef = 0; FRef < VIAMPEG_BUFFERS_MAX; FRef++)
	if (Mpeg->Buffers[FRef].ID == FRefID)
	    break;

    if ((FRef == VIAMPEG_BUFFERS_MAX) || !Mpeg->Buffers[FRef].Mem) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Wrong forward reference buffer (0x%08X).\n",
		   __func__, (unsigned int) FRefID);
	FRef = Buffer; /* this never hurts - at most messes up the video */
	Mpeg->FRefID = 0;
    } else
	Mpeg->FRefID = FRefID;

    /* Check backward reference */
    for (BRef = 0; BRef < VIAMPEG_BUFFERS_MAX; BRef++)
	if (Mpeg->Buffers[BRef].ID == BRefID)
	    break;

    if ((BRef == VIAMPEG_BUFFERS_MAX) || !Mpeg->Buffers[BRef].Mem) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Wrong backward reference buffer (0x%08X).\n",
		   __func__, (unsigned int) BRefID);
	BRef = Buffer; /* this never hurts - at most messes up the video */
	Mpeg->BRefID = 0;
    } else
	Mpeg->BRefID = 0;

    /* Buffers are ready, now prepare for setting up the engine */

    Offset = ((Width << 12) & 0xFFFF0000) | (Width >> 3);

    Picture = 0;
    switch (PictureStructure) {
    case MPEG_PICTURE_FIELD_TOP:
	Picture |= 0x01 << 2;
	break;
    case MPEG_PICTURE_FIELD_BOTTOM:
	Picture |= 0x02 << 2;
	break;
    case MPEG_PICTURE_FRAME:
	Picture |= 0x03 << 2;
	break;
    default:
	break;
    }

    switch (FrameType) {
    case MPEG_FRAMETYPE_I:
	Picture |= 0x01 << 4;
	break;
    case MPEG_FRAMETYPE_P:
	Picture |= 0x02 << 4;
	break;
    case MPEG_FRAMETYPE_B:
	Picture |= 0x03 << 4;
	break;
    default:
	break;
    }

    if (ScanningAlternate)
        Picture |= 1 << 6;

    SliceControl1 = (Mpeg->MBWidth * Mpeg->MBHeight) & 0x3FFF;
    if (PredictionFrame)
	SliceControl1 |= 0x00004000;
    if (TopFieldFirst)
	SliceControl1 |= 0x00008000;
    SliceControl1 |= 0x00010000; /* We are MPEG2 anyway. */
    SliceControl1 |= Mpeg->MBWidth << 18;

     /* calculate (x,y) coordinate of current macroblock */
    SliceControl2 |= (((0x100000 / Mpeg->MBWidth) + 1) & 0xFFFFF) << 4;
    if (IntraHasMV)
	SliceControl2 |= 0x00000001;
    if (UseQScaleTableType1)
	SliceControl2 |= 0x00000002;
    SliceControl2 |= (IntraDCPrecision & 0x03) << 2;
    if (UseIntraDCTable1)
	SliceControl2 |= 0x01000000;

    SliceControl3 = F_MV_Range;
    SliceControl3 |= B_MV_Range << 8;

#if 0
    if (Flags & XumaFlagSecondField)
        SliceControl3 |= 0x00100000;
#endif

    SliceControl3 |= 2 << 16; /* BRef */
    SliceControl3 |= 1 << 18; /* FRef */

    SliceControl3 |= 0xA0 << 16;

    CrOffset = Width * Height;
    CbOffset = CrOffset + (Width >> 1) * (Height >> 1);

#if 0
    ViaDebug(Mpeg->scrnIndex, "  Picture: 0x%08lX\n", Picture);
    ViaDebug(Mpeg->scrnIndex, "  LineOffset: 0x%08lX\n", Offset);
    ViaDebug(Mpeg->scrnIndex, "  SliceControl1: 0x%08lX\n", SliceControl1);
    ViaDebug(Mpeg->scrnIndex, "  SliceControl2: 0x%08lX\n", SliceControl2);
    ViaDebug(Mpeg->scrnIndex, "  SliceControl3: 0x%08lX\n", SliceControl3);
#endif

    /* Set up the actual engine. */
    ViaMpegEngineIdle(Mpeg);
    /* We're not overwriting it yet, so we don't need to check the HQV */

    Mpeg->Regs->Control |= 0x0000001C3;
    Mpeg->Regs->LineOffset = Offset;

    /* Destination buffer */
    Base = Mpeg->Buffers[Buffer].Mem->Base;
    Mpeg->Regs->Buffer[0].Y = Base >> 3;
    Mpeg->Regs->Buffer[0].Cr = (Base + CrOffset) >> 3;
    Mpeg->Regs->Buffer[0].Cb = (Base + CbOffset) >> 3;

    /* Backwards reference */
    Base = Mpeg->Buffers[BRef].Mem->Base;
    Mpeg->Regs->Buffer[1].Y = Base >> 3;
    Mpeg->Regs->Buffer[1].Cr = (Base + CrOffset) >> 3;
    Mpeg->Regs->Buffer[1].Cb = (Base + CbOffset) >> 3;

    /* Forwards reference */
    Base = Mpeg->Buffers[FRef].Mem->Base;
    Mpeg->Regs->Buffer[2].Y = Base >> 3;
    Mpeg->Regs->Buffer[2].Cr = (Base + CrOffset) >> 3;
    Mpeg->Regs->Buffer[2].Cb = (Base + CbOffset) >> 3;

    Mpeg->Regs->Picture = Picture;
    Mpeg->Regs->SliceControl1 = SliceControl1;
    Mpeg->Regs->SliceControl2 = SliceControl2;
    Mpeg->Regs->SliceControl3 = SliceControl3;

    return Success;
}

/*
 *
 */
int
ViaMpegSlicePut(struct ViaMpeg *Mpeg, XID BufferID, CARD8 SliceCode,
                CARD32 SliceLength, CARD8 *Slice)
{
    CARD32 *p = (CARD32 *) Slice, Mask;
    CARD32 i, count;

    MPEGDebug(ViaDebug(Mpeg->scrnIndex, "%s: 0x%08lX: 0x%08lX: %ld bytes\n",
		       __func__, BufferID, SliceCode, SliceLength));

    if (BufferID != Mpeg->CurrentID) {
        xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
                   "%s: failed, 0x%08X is not the current Buffer\n",
		   __func__, (unsigned int) BufferID);
        return BadAccess;
    }

    count = (SliceLength + 3) & ~3; /* round up */
    count += 0x0C; /* add slice code and stop words */

    ViaMpegEngineIdle(Mpeg);

    /* don't overwrite a picture while it's still flipping */
    if (BufferID == Mpeg->FlippingID) {
        ViaHQVIdle(VIAPTR(xf86Screens[Mpeg->scrnIndex])->Swov);
        Mpeg->FlippingID = 0;
    }

    Mpeg->Regs->SliceControl4 = count;

    Mpeg->Regs->SliceStream[0] = 0x00010000 | (SliceCode << 24);

    for (i = (SliceLength >> 2), p = (CARD32 *) Slice; i; i--) {
        Mpeg->Regs->SliceStream[0] = *p;
        p++;
    }

    Mask = (1 << ((SliceLength & 0x03) * 8)) - 1;
    if (Mask)
        Mpeg->Regs->SliceStream[0] = *p & Mask;

    Mpeg->Regs->SliceStream[0] = 0;
    Mpeg->Regs->SliceStream[0] = 0;

    return Success;
}

/*
 *
 */
Bool
ViaMpegBufferIDVerify(struct ViaMpeg *Mpeg, XID BufferID, int Width, int Height)
{
    int i;

    MPEGDebug(ViaDebug(Mpeg->scrnIndex, "%s: 0x%08X (%dx%d)\n",
		       __func__, BufferID, Width, Height));

    if ((Width != Mpeg->Width) || (Height != Mpeg->Height)) {
	xf86DrvMsg(Mpeg->scrnIndex, X_WARNING, "%s: Wrong dimensions for Buffer"
		   " 0x%08X: %dx%d versus %dx%d.\n", __func__,
		   (unsigned int) BufferID, Width, Height,
		   Mpeg->Width, Mpeg->Height);
	return FALSE;
    }

    /* Check ID */
    for (i = 0; i < VIAMPEG_BUFFERS_MAX; i++)
	if (Mpeg->Buffers[i].ID == BufferID)
	    break;

    if (i == VIAMPEG_BUFFERS_MAX) {
        xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Wrong Buffer ID (0x%08X).\n",
		   __func__, (unsigned int) BufferID);
        return FALSE;
    }

    return TRUE;
}

/*
 *
 */
unsigned long
ViaMpegBufferFlip(struct ViaMpeg *Mpeg, XID BufferID)
{
    int i;

    MPEGDebug(ViaDebug(Mpeg->scrnIndex, "%s: 0x%08X\n",
		       __func__, BufferID));

    /* Check ID */
    for (i = 0; i < VIAMPEG_BUFFERS_MAX; i++)
	if (Mpeg->Buffers[i].ID == BufferID)
	    break;

    /* should never happen, as ID should've been checked before. */
    if (i == VIAMPEG_BUFFERS_MAX) {
        xf86DrvMsg(Mpeg->scrnIndex, X_WARNING,
		   "%s: Wrong Buffer ID (0x%08X).\n",
		   __func__, (unsigned int) BufferID);
        return 0;
    }

    if (BufferID == Mpeg->CurrentID)
	ViaMpegEngineIdle(Mpeg);

    Mpeg->FlippingID = BufferID;

    return Mpeg->Buffers[i].Mem->Base;
}

/*
 *
 */
void
ViaMpegBufferFlipRelease(struct ViaMpeg *Mpeg)
{
    VIAFUNC(Mpeg);

    Mpeg->FlippingID = 0;
}

/*
 *
 */
void
ViaMpegInit(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    struct ViaMpeg *Mpeg;

    VIAFUNC(pScrn);

    Mpeg = xnfcalloc(sizeof(struct ViaMpeg), 1);

    Mpeg->scrnIndex = pScrn->scrnIndex;
    Mpeg->Regs = (struct ViaMpegRegs *) (pVia->MapBase + 0xC00);
    pVia->Swov->Mpeg = Mpeg;

    ViaMpegReset(Mpeg);
}

/*
 *
 */
void
ViaMpegStop(ScrnInfoPtr pScrn)
{
    struct ViaMpeg *Mpeg = VIAPTR(pScrn)->Swov->Mpeg;
    int i;

    VIAFUNC(pScrn);

    ViaMpegEngineIdle(Mpeg);
    ViaMpegReset(Mpeg);

    for (i = 0; i < VIAMPEG_BUFFERS_MAX; i++)
        if (Mpeg->Buffers[i].Mem) {
	    Mpeg->Buffers[i].ID = 0;
            ViaMemFree(pScrn, Mpeg->Buffers[i].Mem);
            Mpeg->Buffers[i].Mem = NULL;
        }

    Mpeg->CurrentID = 0;
    Mpeg->FlippingID = 0;
    Mpeg->FRefID = 0;
    Mpeg->BRefID = 0;
}

/*
 *
 */
void
ViaMpegDestroy(ScrnInfoPtr pScrn)
{
    struct ViaSwov *Swov = VIAPTR(pScrn)->Swov;
    struct ViaMpeg *Mpeg = Swov->Mpeg;

    VIAFUNC(pScrn);

    if (!Mpeg)
        return;

    ViaMpegStop(pScrn);

    free(Swov->Mpeg);
    Swov->Mpeg = NULL;
}
