/*
 * Copyright 2004-2009 Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "via_driver.h"

#ifndef _XF86_ANSIC_H
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#endif

#include "via_id.h"
#include "xf86i2c.h"
#include "edid.h"
#include "via_output.h"
#include "via_mode.h"
#include "via_regs.h"

/*
 * Output->Private
 */
struct ViaPanelOutputPrivate {
    int  X;
    int  Y;

#define VIA_DI_12BIT  0x00
#define VIA_DI_24BIT  0x01
    CARD8  BusWidth; /* Digital Output Bus Width */
};

/*
 *
 */
static void
ViaPanelPrivateDestroy(struct ViaOutput *Output)
{
    free(Output->Private);

    /* So we won't try to destroy this twice */
    Output->PrivateDestroy = NULL;
}

/*
 *
 */
static struct ViaPanelOutputPrivate *
ViaPanelPrivateCreate(struct ViaOutput *Output)
{
    VIAFUNC(Output);

    Output->PrivSize = sizeof(struct ViaPanelOutputPrivate);
    Output->Private =  xnfcalloc(1, Output->PrivSize);
    memset(Output->Private, 0, Output->PrivSize);

    Output->PrivateDestroy = ViaPanelPrivateDestroy;

    return Output->Private;
}

/*
 *
 * Option handling.
 *
 */
enum ViaPanelOpts {
    OPTION_BUSWIDTH,
    OPTION_FORCEPANEL,
    OPTION_PANELSIZE
};

static OptionInfoRec ViaPanelOptions[] =
{
#ifdef UNUSED
    {OPTION_BUSWIDTH,	 "BusWidth",  	OPTV_ANYSTR,  {0}, FALSE},
#endif
    {OPTION_FORCEPANEL,  "ForcePanel",  OPTV_BOOLEAN, {0}, FALSE}, /* last resort - don't doc */
    {OPTION_PANELSIZE,   "PanelSize",   OPTV_ANYSTR,  {0}, FALSE},
    {-1,                 NULL,          OPTV_NONE,    {0}, FALSE}
};

/*
 * Stop the user from telling us lies.
 */
/* Scratch register contents. */
#define VIA_PANEL6X4         0
#define VIA_PANEL8X6         1
#define VIA_PANEL10X7        2
#define VIA_PANEL12X7        3
#define VIA_PANEL12X10       4
#define VIA_PANEL14X10       5
#define VIA_PANEL16X12       6
#define VIA_PANEL_INVALID  255

struct {
    int   X;
    int   Y;
    CARD8 Scratch;
} ViaNativePanelSizes[] = {
    { 640,  480, VIA_PANEL6X4  },
    { 800,  600, VIA_PANEL8X6  },
    {1024,  768, VIA_PANEL10X7 },
    {1280,  768, VIA_PANEL12X7 },
    {1280, 1024, VIA_PANEL12X10},
    {1400, 1050, VIA_PANEL14X10},
    {1600, 1200, VIA_PANEL16X12},
    {0, 0, VIA_PANEL_INVALID}
};


/*
 *
 */
static OptionInfoPtr
ViaPanelGetOptions(int scrnIndex, struct ViaPanelOutputPrivate *Private)
{
    OptionInfoPtr  Options;
    char  *s = NULL;

    Options = xnfalloc(sizeof(ViaPanelOptions));
    memcpy(Options, ViaPanelOptions, sizeof(ViaPanelOptions));

    xf86ProcessOptions(scrnIndex, xf86Screens[scrnIndex]->options, Options);

#ifdef UNUSED
    /* Digital Output Bus Width Option */
    Private->BusWidth = VIA_DI_12BIT;
    if ((s = xf86GetOptValString(Options, OPTION_BUSWIDTH))) {
        if (!xf86NameCmp(s, "12BIT")) {
            Private->BusWidth = VIA_DI_12BIT;
            xf86DrvMsg(scrnIndex, X_CONFIG,
                       "Digital Output Bus Width is 12BIT\n");
        } else if (!xf86NameCmp(s, "24BIT")) {
            Private->BusWidth = VIA_DI_24BIT;
            xf86DrvMsg(scrnIndex, X_CONFIG,
                       "Digital Output Bus Width is 24BIT\n");
        }
    }
#endif

    /* ForcePanel is handled in Init, after this function */

    /* Panel Size Option */
    if ((s = xf86GetOptValString(Options, OPTION_PANELSIZE))) {
        int X, Y;
        if (sscanf(s, "%4dx%4d", &X, &Y) == 2) {
            int i;

            for (i = 0; ViaNativePanelSizes[i].X; i++) {
                if ((X == ViaNativePanelSizes[i].X) &&
                    (Y == ViaNativePanelSizes[i].Y)) {
                    Private->X = X;
                    Private->Y = Y;
                    break;
                }
            }

            if (!ViaNativePanelSizes[i].X) {
                xf86DrvMsg(scrnIndex, X_WARNING, "Unknown or Unhandled"
                           " Native Panel Resolution: %dx%d\n", X, Y);
                xf86DrvMsg(scrnIndex, X_INFO,
                           "If this is the actual resolution of the panel, "
                           "contact the driver author.\n");
            }
        } else {
            xf86DrvMsg(scrnIndex, X_WARNING,
                       "Failed to parse PanelSize Option: %s\n", s);
        }
    }

    return Options;
}

/*
 * Only run this during PreInit.
 */
static Bool
ViaPanelNativeResolutionFromConsole(VIAPtr pVia, CARD16 *HDisplay,
				    CARD16 *VDisplay)
{
    CARD8 CR6A, CR6B, CR97, CR99;

    VIAFUNC(pVia);

    if (!xf86IsPrimaryPci(pVia->PciInfo))
	return FALSE; /* VT has nothing set up anyway */

    CR6A = VGACRRead(pVia, 0x6A);
    CR6B = VGACRRead(pVia, 0x6B);
    CR97 = VGACRRead(pVia, 0x97);
    CR99 = VGACRRead(pVia, 0x99);

    if ((/* CRTC2 out of reset and enabled */
	 ((CR6A & 0xC0) == 0xC0) ||
	 /* CRTC2 out of reset but disabled, but simultaneous is enabled */
	 (((CR6A & 0xC0) == 0x40) && (CR6B & 0x08))) &&
	/* DFP source is secondary? */
	(CR97 & 0x10) && (CR99 & 0x10)) {
	/* Use timing of CRTC2 */
	*HDisplay = (VGACRRead(pVia, 0x51) |
		     ((VGACRRead(pVia, 0x55) & 0x70) << 4)) + 1;

	*VDisplay = (VGACRRead(pVia, 0x59) |
		     ((VGACRRead(pVia, 0x5D) & 0x38) << 5)) + 1;

	return TRUE;
    } else if (!(CR97 & 0x10) && !(CR99 & 0x10)) {
	/* Use timing of CRTC1 */
	*HDisplay = (VGACRRead(pVia, 0x01) + 1) * 8;

	*VDisplay = (VGACRRead(pVia, 0x12) |
		     ((VGACRRead(pVia, 0x07) & 0x02) << 7) |
		     ((VGACRRead(pVia, 0x07) & 0x40) << 3) |
		     ((VGACRRead(pVia, 0x35) & 0x04) << 8)) + 1;

	return TRUE;
    } else {
	xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: Unhandled combination:"
		   " CR6A %02X; CR6B %02X; CR97 %02X; CR99 %02X\n",
		   __func__, CR6A, CR6B, CR97, CR99);
	return FALSE;
    }
}

/*
 *
 */
static Bool
ViaPanelNativeResolution(VIAPtr pVia, struct ViaPanelOutputPrivate *Private)
{
    VIAFUNC(pVia);

    /* This could have been provided by config */
    if (!Private->X || !Private->Y) {
        CARD16 HDisplay = 0, VDisplay = 0;
        int i;

        for (i = 0; ViaNativePanelSizes[i].X; i++) {
            if (ViaNativePanelSizes[i].Scratch == pVia->Scratch->PanelSize) {
                Private->X = ViaNativePanelSizes[i].X;
                Private->Y = ViaNativePanelSizes[i].Y;
                break;
            }
        }

        if (!ViaNativePanelSizes[i].X) {
            xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: Unhandled PanelSize"
                       "(%d): Contact the driver author ASAP.\n",
                       __func__, pVia->Scratch->PanelSize);
            Private->X = 1024;
            Private->Y = 768;
        }

        if (ViaPanelNativeResolutionFromConsole(pVia, &HDisplay, &VDisplay)) {
            if ((Private->X != HDisplay) || (Private->Y != VDisplay)) {
                if (!pVia->Scratch->PanelSize) {
                    xf86DrvMsg(pVia->scrnIndex, X_INFO, "%s: Scratch register"
                               " panel sizing was probably wrong.\n", __func__);
                    xf86DrvMsg(pVia->scrnIndex, X_INFO, "%s: Using resolution"
                               " of text-mode panel instead.\n", __func__);
                    Private->X = HDisplay;
                    Private->Y = VDisplay;
                } else
                    xf86DrvMsg(pVia->scrnIndex, X_INFO, "%s: Scratch based"
                               "sizing (%dx%d) doesn't match text-mode "
                               "(%dx%d)\n", __func__, Private->X, Private->Y,
                               HDisplay, VDisplay);
            }
        }
    }

    if (!Private->X || !Private->Y) {
        xf86DrvMsg(pVia->scrnIndex, X_ERROR,
                   "Panel native resolution is unknown.\n");
        return FALSE;
    } else {
        xf86DrvMsg(pVia->scrnIndex, X_INFO,
                   "Panel Native Resolution is %dx%d.\n",
                   Private->X, Private->Y);
        return TRUE;
    }
}

/*
 *
 * Set up all timing for the Panel.
 *
 * We assume that a panel can handle its native resolution at
 * standard CVT timing and at CVT reduced timing.
 *
 */
static void
ViaPanelTimingSet(struct ViaOutput *Output)
{
    struct ViaPanelOutputPrivate *Private = Output->Private;
    DisplayModePtr Modes, Mode;
    float VSync = 60.0;

    VIAFUNC(Output);

    /* To tackle the fact that CVT returns a 29.7H/59.4V mode otherwise */
    if ((Private->X == 640) && (Private->Y == 480))
        VSync = 60.1;

    Output->MonitorName = xnfstrdup("TTL Panel");
    Output->ModesExclusive = FALSE;
    Output->ModesTimingFixed = TRUE;
    Output->ReducedAllowed = TRUE;

    /* Native Reduced mode for fixed timing */
    Mode = ViaCVTMode(Private->X, Private->Y, VSync, TRUE, FALSE);
    Mode->type = M_T_DRIVER | M_T_PREFERRED;
    ViaOutputModesCopyAdd(Output, Mode);
    Modes = Mode;

    /* Native mode for fixed timing */
    Mode = ViaCVTMode(Private->X, Private->Y, VSync, FALSE, FALSE);
    Mode->type = M_T_DRIVER | M_T_PREFERRED;
    ViaOutputModesCopyAdd(Output, Mode);
    Modes->next = Mode;
}

/*
 *
 */
static ModeStatus
ViaPanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
{
    struct ViaPanelOutputPrivate *Private = Output->Private;

    VIAFUNC(Output);

    if ((mode->HDisplay > Private->X) || (mode->VDisplay > Private->Y))
        return MODE_PANEL;

    /* Rest is always caught by Fixed Mode */

    return MODE_OK;
}

/*
 *
 */
static void
ViaPanelMode(struct ViaOutput *Output, DisplayModePtr mode)
{
    ;
}

/*
 * Reduction of old VIA code, untested.
 */
static void
VT3122PanelPower(struct ViaOutput *Output, Bool On)
{
    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");

    /* Enable LCD */
    if (On)
	VGACRMask(Output, 0x6A, 0x08, 0x08);
    else
	VGACRMask(Output, 0x6A, 0x00, 0x08);

    usleep(1);

    if (On) {
        VGACRMask(Output, 0x91, 0x10, 0x10);
        usleep(0x19);
        VGACRMask(Output, 0x91, 0x08, 0x08);
        usleep(0x1FE);
        VGACRMask(Output, 0x91, 0x06, 0x06);
        usleep(0x01);
    } else {
        VGACRMask(Output, 0x91, 0x00, 0x06);
        usleep(0xD2);
        VGACRMask(Output, 0x91, 0x00, 0x08);
        usleep(0x19);
        VGACRMask(Output, 0x91, 0x00, 0x10);
        usleep(0x01);
    }
}

/*
 * Reduction of old VIA code, untested.
 */
static void
VT7205PanelPower(struct ViaOutput *Output, Bool On)
{
    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");

    /* Enable LCD */
    if (On)
	VGACRMask(Output, 0x6A, 0x08, 0x08);
    else
	VGACRMask(Output, 0x6A, 0x00, 0x08);

    usleep(1);

    if (On) {
        VGACRMask(Output, 0x91, 0x10, 0x10);
        usleep(0x19);
        VGACRMask(Output, 0x91, 0x08, 0x08);
        usleep(0x1FE);
        VGASRMask(Output, 0x3D, 0x20, 0x20);
        usleep(0x01);
    } else {
        VGASRMask(Output, 0x3D, 0x00, 0x20);
        usleep(0xD2);
        VGACRMask(Output, 0x91, 0x00, 0x08);
        usleep(0x19);
        VGACRMask(Output, 0x91, 0x00, 0x10);
        usleep(0x01);
    }

    usleep(1);
}

/*
 * This implements only superficial power down. Good enough for DPMS.
 */
static void
VT3108PanelPower(struct ViaOutput *Output, Bool On)
{
    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");

    VGACRMask(Output, 0x91, 0x00, 0x01); /* Disable Software Control */
    usleep(1);

    if (On) /* Direct Panel Display and Backlight control */
        VGACRMask(Output, 0x91, 0x00, 0xC0);
    else
        VGACRMask(Output, 0x91, 0xC0, 0xC0);
}

/*
 * future: I2CDevPtr for LVDS encoders.
 */
struct ViaOutput *
ViaPanelInit(int scrnIndex, I2CDevPtr pDev)
{
    VIAPtr pVia = VIAPTR(xf86Screens[scrnIndex]);
    struct ViaOutput *Output;
    struct ViaPanelOutputPrivate *Private;

    VIAFUNC(pVia);

    if (pDev)
        xf86DrvMsg(scrnIndex, X_WARNING, "%s: Ignoring I2C Device"
                   " passed.\n", __func__);

    Output = xnfcalloc(1, sizeof(struct ViaOutput));

    Output->Prev = NULL;
    Output->Next = NULL;
    Output->scrnIndex = scrnIndex;
    Output->I2CDev = pDev;
    Output->Type = OUTPUT_PANEL;

    Private = ViaPanelPrivateCreate(Output);

    /* Parse options here ourselves. */
    Output->Options = ViaPanelGetOptions(scrnIndex, Private);

    /* Is there even a panel present? */
    if (xf86ReturnOptValBool(Output->Options, OPTION_FORCEPANEL, FALSE)) {
        xf86DrvMsg(scrnIndex, X_INFO, "Enabling panel from config.\n");
        xf86DrvMsg(scrnIndex, X_WARNING,
                   "This is a last resort option only.\n");
        xf86DrvMsg(scrnIndex, X_WARNING,
                   "Report PCI IDs to http://unichrome.sf.net/ ASAP.\n");
    } else if (pVia->Id && pVia->Id->HasPanel) {
        xf86DrvMsg(scrnIndex, X_INFO, "Enabling panel from"
                   " PCI-Subsystem Id information.\n");
    } else {
        Output->PrivateDestroy(Output);

        free(Output->Options);
        free(Output);
        return NULL;
    }

    if (!ViaPanelNativeResolution(pVia, Private)) {
        xf86DrvMsg(scrnIndex, X_ERROR, "%s: Disabling panel.\n", __func__);
        Output->PrivateDestroy(Output);
        free(Output->Options);
        free(Output);
        return NULL;
    }

    Output->Name = "Panel";

    Output->ClockMaster = FALSE;

    ViaPanelTimingSet(Output);

    Output->Save = NULL;
    Output->Restore = NULL;
    Output->Sense = NULL;
    Output->ModeValid = ViaPanelModeValid;
    Output->Mode = ViaPanelMode;

    switch (pVia->Chipset) {
    case VT3122:
        Output->Power = VT3122PanelPower;
        break;
    case VT7205:
        Output->Power = VT7205PanelPower;
        break;
    case VT3108:
    case VT3344:
        Output->Power = VT3108PanelPower;
        break;
    default:
	xf86DrvMsg(scrnIndex, X_WARNING, "%s: Unsupported chipset.\n", __func__);
        Output->PrivateDestroy(Output);
        free(Output->Options);
        free(Output);
        return NULL;
    }

    return Output;
}
